@@ -82,18 +82,19 @@ static char *get_exports(struct autofs_point *ap, const char *host)
 {
 	char buf[MAX_ERR_BUF];
 	char *mapent;
-	exports exp;
+	exports exp, this;
 
 	debug(ap->logopt, MODPREFIX "fetchng export list for %s", host);
 
 	exp = rpc_get_exports(host, 10, 0, RPC_CLOSE_NOLINGER);
 
 	mapent = NULL;
-	while (exp) {
+	this = exp;
+	while (this) {
 		if (mapent) {
 			int len = strlen(mapent) + 1;
 
-			len += strlen(host) + 2*(strlen(exp->ex_dir) + 2) + 3;
+			len += strlen(host) + 2*(strlen(this->ex_dir) + 2) + 3;
 			mapent = realloc(mapent, len);
 			if (!mapent) {
 				char *estr;
@@ -103,10 +104,10 @@ static char *get_exports(struct autofs_point *ap, const char *host)
 				return NULL;
 			}
 			strcat(mapent, " \"");
-			strcat(mapent, exp->ex_dir);
+			strcat(mapent, this->ex_dir);
 			strcat(mapent, "\"");
 		} else {
-			int len = 2*(strlen(exp->ex_dir) + 2) + strlen(host) + 3;
+			int len = 2*(strlen(this->ex_dir) + 2) + strlen(host) + 3;
 
 			mapent = malloc(len);
 			if (!mapent) {
@@ -117,16 +118,16 @@ static char *get_exports(struct autofs_point *ap, const char *host)
 				return NULL;
 			}
 			strcpy(mapent, "\"");
-			strcat(mapent, exp->ex_dir);
+			strcat(mapent, this->ex_dir);
 			strcat(mapent, "\"");
 		}
 		strcat(mapent, " \"");
 		strcat(mapent, host);
 		strcat(mapent, ":");
-		strcat(mapent, exp->ex_dir);
+		strcat(mapent, this->ex_dir);
 		strcat(mapent, "\"");
 
-		exp = exp->ex_next;
+		this = this->ex_next;
 	}
 	rpc_exports_free(exp);
 
@@ -148,19 +149,13 @@ static int do_parse_mount(struct autofs_point *ap, struct map_source *source,
 	ret = ctxt->parse->parse_mount(ap, name, name_len,
 				 mapent, ctxt->parse->context);
 	if (ret) {
-		time_t now = time(NULL);
 		struct mapent_cache *mc = source->mc;
-		struct mapent *me;
-		int rv = CHE_OK;
 
+		/* Don't update negative cache when re-connecting */
+		if (ap->flags & MOUNT_FLAG_REMOUNT)
+			return NSS_STATUS_TRYAGAIN;
 		cache_writelock(mc);
-		me = cache_lookup_distinct(mc, name);
-		if (!me)
-			rv = cache_update(mc, source, name, NULL, now);
-		if (rv != CHE_FAIL) {
-			me = cache_lookup_distinct(mc, name);
-			me->status = now + ap->negative_timeout;
-		}
+		cache_update_negative(mc, source, name, ap->negative_timeout);
 		cache_unlock(mc);
 		return NSS_STATUS_TRYAGAIN;
 	}
@@ -210,7 +205,7 @@ next:
 			goto cont;
 
 		debug(ap->logopt, MODPREFIX
-		      "attempt to update exports for exports for %s", me->key);
+		      "attempt to update exports for %s", me->key);
 
 		master_source_current_wait(ap->entry);
 		ap->entry->current = source;
@@ -315,8 +310,10 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 				cache_writelock(smc);
 				sme = cache_lookup_distinct(smc, name);
 				/* Negative timeout expired for non-existent entry. */
-				if (sme && !sme->mapent)
-					cache_delete(smc, name);
+				if (sme && !sme->mapent) {
+					if (cache_pop_mapent(sme) == CHE_FAIL)
+						cache_delete(smc, name);
+				}
 				cache_unlock(smc);
 			}
 		}
